/*    Copyright (C) 2005 XIAO, Gang of Universite de Nice - Sophia Antipolis
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

	/* smkd configuration macro processing. */

/* Prefix processing */

char prefixbuf[PREFIXBUFSIZE];
struct {
    unsigned char *name;
    unsigned char *def;
    int hash;
    short int namelen, deflen;
} prefixes[PREFIXLIM];
int prefixend[256];
int prefixcnt;
time_t prefixftime;

void getprefixes(void)
{
    struct stat st;
    char *p1, *p2, *p3;
    int i, h;

    if(stat(prefixfile,&st)) {	/* passwd file not found */
	prefixcnt=0; return;
    }
	/* Fix the access rights if necessary */
    if(denyid && st.st_uid!=denyid) chown(prefixfile,denyid,denyid);
    if(prefixftime!=0 && (st.st_mtime==prefixftime || st.st_mtime==now)) return;	/* not modified */
    prefixftime=st.st_mtime;
    _readfile(prefixfile,prefixbuf,sizeof(prefixbuf));
    prefixcnt=0;
    for(p1=find_word_start(prefixbuf);
	*p1 && prefixcnt < PREFIXLIM; p1=find_word_start(p2)) {
	p2=find_line_end(p1); if(*p2) *p2++=0;
	if(*p1!='/') continue;
	p1++;
	p3=find_word_end(p1); if(*p3==0) continue;
	*p3++=0; p3=find_word_start(p3); if(*p3!='/') continue;
	p3++; *find_word_end(p3)=0;
	if(p3-p1 > PREFIXNAMELIM) continue;
	i=search_list(prefixes,prefixcnt,sizeof(prefixes[0]),p1);
		/* multiple definitions: later ones are ignored. */
	if(i<0) {
	    i=~i;
	    memmove(prefixes+i+1,prefixes+i,(prefixcnt-i)*sizeof(prefixes[0]));
	    prefixes[i].name=p1;
	    prefixes[i].def=p3;
	    prefixes[i].namelen=strlen(p1);
	    prefixes[i].deflen=strlen(p3);
	    prefixes[i].hash=strhash(p1);
	    prefixcnt++;
	}
    }
    for(i=0;i<256;i++) prefixend[i]=-1;
    for(i=0;i<prefixcnt;i++) {
	h=prefixes[i].name[0]; prefixend[h]=i;
    }
}

	/* Returns 0 if OK. */
int prefix_translate(char *p, int hash, int lim)
{
    int first;
    int i, l;

    newhash=hash;
    if(*p!='/') return 0;
    p++; first=*p;
    if(prefixend[first]<0) return 0;
    for(i=prefixend[first]; i>=0 && prefixes[i].name[0]==first; i--) {
	if((~hash) & prefixes[i].hash) continue;
	if(strncmp(p,prefixes[i].name,prefixes[i].namelen)!=0) continue;
	l=strlen(p)+1;
	if(l+prefixes[i].deflen-prefixes[i].namelen >= lim) return ENAMETOOLONG;
	memmove(p+prefixes[i].deflen,p+prefixes[i].namelen,l-prefixes[i].namelen);
	memmove(p,prefixes[i].def,prefixes[i].deflen);
	newhash=strhash(p);
	return 0;
    }
    return 0;
}

/* macro processing */

char macrobuf[MACROLEN];
struct {
    char *name;
    char *def;
} macros[MACROLIM];

int macrocnt;
int macroloopcnt;
time_t macroftime;

 	/* Points to the end of a line */
char *macro_line_end(char *p)
{
    char *pp;
    
    for(pp=strchr(p,'\n'); pp>p && pp[-1]=='\\'; pp=strchr(++pp,'\n'));
    if(pp==NULL) pp=p+strlen(p);
    return pp;
}

void macro_unescape_line(char *p)
{
    char *pp;
    
    for(pp=strchr(p,'\\'); pp; pp=strchr(++pp,'\\')) {
	if(pp[1]=='\n') *pp=' ';
    }
}

void getmacros(void)
{
    struct stat st;
    char *p1, *p2, *p3;
    int i;

    getprefixes();
    if(stat(macrofile,&st)) {	/* passwd file not found */
	macrocnt=0; return;
    }
	/* Fix the access rights if necessary */
    if(denyid && st.st_uid!=denyid) chown(macrofile,denyid,denyid);
    if(macroftime!=0 && (st.st_mtime==macroftime || st.st_mtime==now)) return;	/* not modified */
    macroftime=st.st_mtime;
    confcache_init();
    _readfile(macrofile,macrobuf,sizeof(macrobuf));
    macrocnt=0;
    for(p1=find_word_start(macrobuf);
	*p1 && macrocnt < MACROLIM; p1=find_word_start(p2)) {
	p2=macro_line_end(p1); if(*p2) *p2++=0;
	if(!myisalnum(*p1)) continue;
	p3=strchr(p1,'='); if(!p3) continue;
	if(p3-p1 > MACRONAMELIM) continue;
	*p3++=0;
	macro_unescape_line(p3);
	p3=find_word_start(p3);
	strip_trailing_spaces(p3);
	strip_trailing_spaces(p1);
	i=search_list(macros,macrocnt,sizeof(macros[0]),p1);
		/* multiple definitions: newer one covers earlier ones */
	if(i<0) {
	    i=~i;
	    memmove(macros+i+1,macros+i,(macrocnt-i)*sizeof(macros[0]));
	    macros[i].name=p1;
	    macrocnt++;
	}
	macros[i].def=p3;
    }
}

	/* macro substitution. */
void msubstit(int token, char *p, size_t lim)
{
    char *pp, *p1, *p2, *p3, *ps;
    char nbuf[MACRONAMELIM];
    char tbuf[64];
    int l, ln, l2, i;

    for(pp=mystrstr(p,"%("); pp!=NULL; pp=mystrstr(pp,"%(")) {
	p1=pp+2; i=0;
	for(p2=p1; *p2 && (myisalnum(*p2) || strchr(" 	_-%()",*p2)!=NULL); p2++) {
	    if(*p2=='(') i++;
	    if(*p2==')') {
		i--; if(i<0) break;
	    }
	}
	if(*p2!=')') {
	    pp++; continue;
	}
	if(p2-p1<=0 || p2-p1>=MACRONAMELIM) continue;
	macroloopcnt++;
	if(macroloopcnt > MACROLOOPLIM) {
	    logprefix("ERROR msubstit: too many nested substitutions.\n");
	    continue;
	}
	memmove(nbuf,p1,p2-p1); nbuf[p2-p1]=0;
	msubstit(token,nbuf,sizeof(nbuf));
	*p2++=0; ln=strlen(p2)+1;
	/* p3 points to possible macro parameters */
	p3=find_word_end(nbuf);
	if(*p3) *p3++=0;
	if(strcmp(nbuf,"token")==0) {
	    if(token&0xFFFF0000) tbuf[0]='u'; else tbuf[0]='s';
	    int2str(token&0xFFFF,tbuf+1,sizeof(tbuf)-2);
	    ps=tbuf;
	}
	else {
	    i=search_list(macros,macrocnt,sizeof(macros[0]),nbuf);
	    if(i<0) {
		if(wordchr(p3,"maymiss")==NULL) {
		    logprefix("ERROR msubstit: unknown macro ");
		    print(p1); print(".\n");
		}
		memmove(pp,p2,ln); continue;
	    }
	    ps=macros[i].def;
	}
	l2=strlen(ps);
	l=(pp-p) + ln + l2;
	if(l >= lim) {
	    logprefix("ERROR msubstit: string too long.\n");
	    *p=0; return;
	}
	memmove(pp+l2,p2,ln);
	memmove(pp,ps,l2);
    }
}

